function [selF, delF, time] = icf(data)
% Implementation of the Incremental Cross-correlation Filter (ICF)
% algorithm for unspervised feature selection in multivariate sensor
% timeseries.
%
% [selF, delF, time] = icf(data)
%
% Performs the ICF algorithm on the collection of multivariate timeseries
% in "data" and returns the orginal indices of the selected and deleted 
% features in 'selF' and 'delF', respectively.  
% 
% Inputs are:
%
% data - a {1xN} cell array containing N multivariate timeseries,
% represented as MxV arrays of doubles, where V is the number of features
% (i.e. univariate timeseries) and M is the length of the multivariate
% timeseries, which varies between the N samples.
%
% The fucntion returns: 
%
% selF -> (1xV1) array containing the indices of the V1 features selected 
% by the algorithm (out of the initial V features).
% 
% delF -> (1xV2) array containing the indices of the V2 features selected 
% by the algorithm (out of the initial V features). Note that V = V1+V2.
%
% time -> (1x1) the time required to run the algorithm. 
%
% The code in this function implements the ICF algorithm described in: 
% 
% D. Bacciu. "Unsupervised feature selection for sensor time-series 
% in pervasive computing applications." Neural Computing and Applications 
% (2015): 1-15.
% 
% Please cite the article if you use this code and inform the author if you
% intend to redistribute the code.
% 
% ICF Algorithm V1.0
% Written by Davide Bacciu (2015) 
% Computer Science Department - University of Pisa
% Email: bacciu@di.unipi.it
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The MIT License (MIT)
% 
% Copyright (c) [2015] [Davide Bacciu]
% 
% Permission is hereby granted, free of charge, to any person obtaining a copy
% of this software and associated documentation files (the "Software"), to deal
% in the Software without restriction, including without limitation the rights
% to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
% copies of the Software, and to permit persons to whom the Software is
% furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included in all
% copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
% IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
% AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
% OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
% SOFTWARE.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

time=cputime;
%%%%%%%%%%%%%%%%% INITIALIZATIONS
%Threshold used to define what an high correlation is
TH_CORR = 0.99; 
%Threshold used to define what a low autocorrelation is
TH_AUTO = 0.1;  %For mean
TH_AUTO_VAR = 0.001; %For variance
%Threshold used to define the percentage of highly correlated sequences
%allowed
TH_SEQP = 0.2;

%Number of inputs
Ninput = size(data{1},2);
%Number of experiments
Nseq = length(data);

%Init the counter matrix
mask = zeros(Ninput,Ninput);

%%%%%%%%%%%%%%%%% CORRELATION MASK COMPUTATION

%First compute cross correlation using the helper function
[crossAll,autocorr] = xcorrHelp(data);

%Compute number of experiments in which we have high correlation
for i = 1:Nseq,
    mask = mask + double(crossAll(:,:,i)>= TH_CORR); %Count superlarge correlations
end;

%At least TH_SEQP * 100 sequences have a correlation value higher than
%TH_CORR for a feature
mask = mask / Nseq;
maskN = (mask) > TH_SEQP;

%Zero the diagonal
for i = 1:Ninput,
    maskN(i,i) = false;
end
 
%%%%%%%%%%%%%%%%% ITERATIVE FILTER

%Initially all features are selected
X = maskN;
selF = [];
delF = [];

%Conversion from original input space to X indices
fromOr2X = (1:Ninput);
fromX2Or = (1:Ninput);

%First perform auto-correlation check to delete the noisy features
todelX =  (mean(autocorr,1) < TH_AUTO) & (var(autocorr,[],1) < TH_AUTO_VAR);
%Save ORIGINAL indices of deleted elements
[delF, X, fromOr2X, fromX2Or] = update_deleted(todelX,delF, X, fromOr2X, fromX2Or);

notdone = true;
while notdone
    
    %This will prevent deadlock
    notdone = false;
    
    %RULE 0 - Check if any of the rows in X is uncorrelated with everybody
    %else and save it
    toKeepX = (sum(X,2) == 0)';
    if (sum(toKeepX) > 0),
        %Save ORIGINAL indices of selected elements
        [selF, X, fromOr2X, fromX2Or] = update_deleted(toKeepX, selF, X, fromOr2X, fromX2Or);
        
        %Make sure that any feature that now becomes uncorrelated is
        %cancelled (THEY WHERE CORRELATED WITH THE CANCELLED ONES)
        todelX = (sum(X,2) == 0)';
        if (sum(todelX) > 0),
            %Save ORIGINAL indices of selected elements
            [delF, X, fromOr2X, fromX2Or] = update_deleted(todelX,delF, X, fromOr2X, fromX2Or);
        end 
    end %RULE 0
    
    %Break the while if finished
    if (isempty(X))
        break;
    end
    
    %RULE 1 - Check if any of the rows in X is correlated with everybody
    %else and delete it
    todelX =  (sum(X,2) == (size(X,2)-1))';
    if (sum(todelX) > 0) && (sum(todelX) < size(X,2)), %Not all the features
        %Save ORIGINAL indices of deleted elements
        [delF, X, fromOr2X, fromX2Or] = update_deleted(todelX,delF, X, fromOr2X, fromX2Or);
        %Allow to do an additional iteration
        notdone = true;
    else
        %RULE 2 - if all features left are mutually correlated let
        %live THE ONE less correlated with the features selected so
        %far and terminate
        if (sum(todelX) ==  (size(X,2))), %All the features
            if (~isempty(selF))
                %If the features are originally correlated with the selected
                %one, can be deleted.
                corrSelF = sum(maskN(selF,fromX2Or(todelX)),1) > 0;
                if  (sum(corrSelF) == length(corrSelF)),
                    %Delete all
                    delF = [delF fromX2Or(todelX)];
                    X = [];
                    fromOr2X = [];
                    fromX2Or = [];
                else
                    %Compute the correlation coefficients with respect to
                    %selected
                    corrCoeff = mean(mask(:,selF),2); %SUM?
                    %Select the minimum correlated in X space
                    [mn,mnind] = min(corrCoeff(fromX2Or(todelX)));
                    %Save the minimum correlated in OR space to selF
                    selF = [selF fromX2Or(mnind)];
                    %Put the rest in delF
                    todelX(mnind) = false;
                    delF = [delF fromX2Or(todelX)];
                    %Delete the rest
                    X = [];
                    fromOr2X = [];
                    fromX2Or = [];
                end
            end
            %No more iterations needed
            notdone = false;
            
            if (isempty(selF)) %May also act as an else to the previous IF
                %PROTECTIVE CHOICE: SAVE ANYTHING LEFT FOR WRAPPER PHASE
                disp('Protective'); 
                mx =  min(mask);
                mx1 = min(mx);
                toKeepX = (mx == mx1);     
                toKeepX(delF) = []; %Get rid of potentially deleted vars
                %Save ORIGINAL indices of selected elements
                [selF, X, fromOr2X, fromX2Or] = update_deleted(toKeepX, selF, X, fromOr2X, fromX2Or);
                %Allow another iteration
                notdone = true;                
            end
          
        else % end RULE 2
            
            %RULE 3 - Extract the rows "i" in X correlated with a minimum number of
            %elements "j" and delete j it if is one of the maximum correlated
            %features 
            tot = sum(X,2);
            [mn,mnind] = min(tot);
            toKeepX = (tot == mn)';
            if (sum(toKeepX) == length(toKeepX)) %They are all equally correlated
                %Just delete the highly correlated with the saved Features
                if (~isempty(selF))                    
                    %Compute the correlation coefficients with respect to
                    %selected
                    corrCoeff = mean(mask(:,selF),2); %SUM?
                    %Select the maximum correlated in X space
                    [mx,mxind] = max(corrCoeff(fromX2Or(toKeepX)));
                    %Save ORIGINAL indices of deleted elements
                    toDelJ = false(size(fromX2Or));
                    toDelJ(mxind) = true;
                    [delF, X, fromOr2X, fromX2Or] = update_deleted(toDelJ, delF, X, fromOr2X, fromX2Or);                                      
                end
            else
                for i=find(toKeepX),
                    %Find j with maximum number of correlated featers
                    mx = max(tot(X(i,:)));
                    j = find((tot' == mx) & X(i,:));
                    if (tot(j) > tot(i))
                        %Save ORIGINAL indices of deleted elements
                        toDelJ = false(size(fromX2Or));
                        toDelJ(j) = true;
                        [delF, X, fromOr2X, fromX2Or] = update_deleted(toDelJ, delF, X, fromOr2X, fromX2Or);
                        %Allow to do an additional iteration
                        notdone = true;
                        %Breaks the "for i" cycle to avoid inconsistencies
                        break;
                    else
                        if (~isempty(selF))
                            corrCoeff = mean(mask(:,selF),2); %SUM?
                            if (corrCoeff(fromX2Or(j)) >= corrCoeff(fromX2Or(i)))
                                %Save ORIGINAL indices of deleted elements
                                toDelJ = false(size(fromX2Or));
                                toDelJ(j) = true;
                            else
                                %Save ORIGINAL indices of deleted elements
                                toDelJ = false(size(fromX2Or));
                                toDelJ(i) = true;
                            end
                            [delF, X, fromOr2X, fromX2Or] = update_deleted(toDelJ, delF, X, fromOr2X, fromX2Or);
                            %Breaks the "for i" cycle to avoid inconsistencies
                            break;
                        else %(isempty(selF),
                            %PROTECTIVE CHOICE: UNLIKELY TO HAPPEN
                            disp('Protective 2');
                            toKeepX =  false(size(fromX2Or));
                            toKeepX([i j]) = true;
                            [selF, X, fromOr2X, fromX2Or] = update_deleted(toKeepX, selF, X, fromOr2X, fromX2Or);
                            break;
                        end
                    end
                end %FOR  RULE 3
            end 
            %Allow to do an additional iteration
            notdone = true;
        end      %RULE 3
    end %RULE 1
end %WHILE

time = cputime-time;
end %MAIN FUNCTION


%Helper function to maintain updated structure
function [delF, X, fromOr2X, fromX2Or] = update_deleted(todelX,delF, X, fromOr2X, fromX2Or)
%Save ORIGINAL indices of selected/deleted elements
delF = [delF fromX2Or(todelX)];
%Delete features from X
X(todelX,:)=[]; %row
X(:,todelX)=[]; %cols
%Update indices
fromOr2X((~todelX)) = 1:size(X,1);
fromOr2X(fromX2Or(todelX)) = 0;
fromX2Or(todelX) = [];
end

%Helper function to compute cross correlation on sequences
function [corrData,autocorr] = xcorrHelp(dataset)
%Ninputs
Ninput = size(dataset{1},2);

%XCorrelation matrix
corrData = zeros(Ninput,Ninput,length(dataset));

%for each experiment
for j = 1:length(dataset),          
        %Extract data 
        data =  dataset{j};
        
        %Normalized cross corr \phi_xy(t) / sqrt(\phi_xx(0) * \phi_yy(0))
        crosscor = xcorr(data,'coeff');
                  
        %Mean subtract
        data = data - repmat(mean(data,1),[size(data,1) 1]);
        %Biased xcorrelation for autocorrelation 
        crosscor1 = xcorr(data,'biased');

        %Take maximum (i.e. maximally correlated instant in time)
        corrData(:,:,j) = reshape(max(abs(crosscor),[],1), [size(data,2) size(data,2) 1]);
        
        %Compute off time 0 mean autocorrelation (White noise has ideally 0
        %autocorrelation off time 0)
        ind = [1:(size(data,1)-1) (size(data,1)+1):size(crosscor1,1)];
        autocorr(j,:) = diag(reshape(mean(abs(crosscor1(ind,:)),1), [size(data,2) size(data,2) 1]));
end
end

